import torch
import torch.nn as nn
import torch.nn.functional as F
from typing import Callable, Tuple, Union
import warnings

from . import backend


class Signature(nn.Module):
    """Module wrapper around the signatory.signature function. See ``signatory.signature.__doc__``.

    Arguments:
        depth: as ``signatory.signature.__doc__``.

        basepoint: as ``signatory.signature.__doc__``.

        stream: as ``signatory.signature.__doc__``.

        flatten: as ``signatory.signature.__doc__``.

    Called with a single argument ``path`` of type ``torch.Tensor``.
    """

    def __init__(self, depth: int, basepoint: bool = False, stream: bool = False, flatten: bool = True, **kwargs):
        if not isinstance(depth, int) or depth < 1:
            raise ValueError('Depth must be an integer greater than or equal to one. Given {depth} of type '
                             '{tdepth}'.format(depth=depth, tdepth=type(depth)))
        super(Signature, self).__init__(**kwargs)
        self.depth = depth
        self.basepoint = basepoint
        self.stream = stream
        self.flatten = flatten

    def forward(self, path: torch.Tensor) -> Union[torch.Tensor, Tuple[torch.Tensor]]:
        if path.size(1) == 1:
            warnings.warn('{clsname} called on path with only one channel; the signature is now just the moments of the'
                          ' path, so there is no interesting information from cross terms.'
                          .format(clsname=self.__class__.__name__))
        return backend.signature(path, self.depth, self.basepoint, self.stream, self.flatten)

    def extra_repr(self):
        return 'depth={depth}'.format(depth=self.depth)


class Augment(nn.Module):
    r"""Augmenting the stream before feeding it into a signature is crucial to obtain higher-order information in the
    signature. One way to do this is in a data-dependent way is to apply a feedforward neural network to sections of the
    stream, so as to obtain another stream; on this stream the signature is then applied.

    The input path is expected to be a three-dimensional tensor, with dimensions :math:`(N, C, L)`, where
    :math:`N` is the batch size, :math:`C` denotes the number of channels, and :math:`L` is the length of the input
    sequence. Thus each batch element is interpreted as a stream of data :math:`(x_1, \ldots, x_L)`, where each
    :math:`x_i \in \mathbb{R}^C`. (This is the same as ``torch.nn.Conv1D``, for example.)

    Then this stream may be 'augmented' via some function

    .. math::
        \phi \colon \mathbb{R}^{C \times k} \to \mathbb{R}^{\widehat{C}}

    giving a stream of data

    .. math::
        \left(\phi(x_1, ... x_k), \ldots, \phi(x_{n - k + 1}, \ldots, x_n)\right),

    which is essentially a three-dimensional tensor with dimensions :math:`(N, \widehat{C}, L - k + 1)`.

    Thus this essentially operates as a one dimensional convolution, except that a whole network is swept across the
    input, rather than just a single convolutional layer.

    Both the original stream and time can be specifically included in the augmentation. (This usually tends to give
    better empirical results.) For example, if both :attr:`include_original` is True and :attr:`include_time` is True,
    then each :math:`\phi(x_i, ... x_{k + i - 1})` is of the form

    .. math::
        \left(\frac{i}{T}, x_i, \varphi(x_i, ... x_{k + i - 1})\right).

    where :math:`T` is a constant appropriately chosen so that the first entry moves between :math:`0` and :math:`1` as
    :math:`i` varies. (Specifically, :math:`T = L - k + 1 + 2 \times \text{padding}`)

    For further details see `Deep Signatures -- Bonnier et al. 2019 <https://arxiv.org/abs/1905.08494>`_. (Thus this
    module is here more for convenience: it doesn't directly relate to the signature transform; it's just useful to have
    around when you are using the signature transform.)

    Arguments:
        in_channels (int): Number of channels :math:`C` in the input stream.

        layer_sizes (tuple of int): Specifies the sizes of the layers of the feedforward neural network to apply to
            the stream. The final value of this tuple specifies the number of channels in the augmented stream,
            corresponding to the value :math:`\widehat{C}` in the preceding discussion.

        kernel_size (int): Specifies the size of the kernel to slide over the stream, corresponding to the value
            :math:`k` in the preceding discussion.

        stride (int, optional): Defaults to 1. How far to move along the input stream before re-applying the
            feedforward neural network. Thus the output stream is given by

            .. math::
                (\phi(x_1, \ldots, x_k),
                 \phi(x_{1 + \text{stride}}, \ldots, x_{k + 2 \times \text{stride}}),
                 \phi(x_{1 + 2 \times \text{stride}}, \ldots, x_{k + 2 \times \text{stride}}),
                 \ldots)

        padding (int, optional): Defaults to 0. How much zero padding to add to either end of the the input stream
            before sweeping the feedforward neural network along it.

        dilation (int, optional): The spacing between input elements given to the feedforward neural network.
            Defaults to 1. Harder to describe; see the equivalent argument for ``torch.nn.Conv1D``.

        bias (bool, optional): Defaults to True. Whether to use biases in the neural network.

        activation (callable, optional): Defaults to ReLU. The activation function to use in the feedforward neural
            network.

        include_original (bool, optional): Defaults to True. Whether or not to include the original stream
            (pre-augmentation) in the augmented stream.

        include_time (bool, optional): Defaults to True. Whether or not to also augment the stream with a 'time' value.
            These are values in :math:`[0, 1]` corresponding to how far along the stream dimension the element is.

    .. note::

        Thus the resulting stream of data has size ``out_channel``, where in pseudocode:

        .. code-block:: python

            out_channel = layer_sizes[-1]
            if include_original:
                out_channel += in_channels
            if include_time:
                out_channel += 1
    """

    def __init__(self, in_channels: int, layer_sizes: Tuple[int], kernel_size: int, stride: int = 1, padding: int = 0,
                 dilation: int = 1, bias: bool = True, activation: Callable[[torch.Tensor], torch.Tensor] = F.relu,
                 include_original: bool = True, include_time: bool = True, **kwargs):
        super(Augment, self).__init__(**kwargs)

        if isinstance(layer_sizes, int):
            layer_sizes = (layer_sizes,)

        self.in_channels = in_channels
        self.layer_sizes = layer_sizes
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.dilation = dilation
        self.bias = bias
        self.activation = activation
        self.include_original = include_original
        self.include_time = include_time

        self.convs = nn.ModuleList()
        if layer_sizes:
            self.convs.append(nn.Conv1d(in_channels=in_channels,
                                        out_channels=layer_sizes[0],
                                        kernel_size=kernel_size,
                                        stride=stride,
                                        padding=padding,
                                        dilation=dilation,
                                        bias=bias))
            last_layer_channels = layer_sizes[0]
            for augment_channel in layer_sizes[1:]:
                # These pointwise convolutions correspond to sliding a standard feedforward network across the input.
                self.convs.append(nn.Conv1d(in_channels=last_layer_channels,
                                            out_channels=augment_channel,
                                            kernel_size=1,
                                            bias=bias))
                last_layer_channels = augment_channel

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if len(x.shape) != 3:
            raise RuntimeError('Argument x should have three dimensions, (batch, channnel, stream). Given shape'
                               '{shape} dimensions with {x}.'.format(shape=x.shape, x=x))
        pieces = []
        if self.include_original:
            truncated_x = x.narrow(2, self.kernel_size - 1, x.size(2) - self.kernel_size + 1)
            pieces.append(truncated_x)

        if self.include_time:
            time = torch.linspace(0, 1, x.size(2) - self.kernel_size + 1, dtype=torch.float, device=x.device)
            time = time.expand(x.size(0), 1, -1)
            pieces.append(time)

        if self.layer_sizes:
            augmented_x = self.convs[0](x)
            for conv in self.convs[1:]:
                augmented_x = self.activation(augmented_x)
                augmented_x = conv(augmented_x)
            pieces.append(augmented_x)
        return torch.cat(pieces, dim=1)  # concatenate along channel axis

    def extra_repr(self):
        return ('activation={activation}, include_original={include_original}, include_time={include_time}'
                ).format(activation=self.activation, include_original=self.include_original,
                         include_time=self.include_time)



import torch  # must be imported before anything from signatory

from .backend import (signature,
                      signature_channels)
from .modules import (Signature,
                      Augment)


__version__ = "0.1.0"

del torch



import torch
import torch.autograd as autograd
from typing import Any, Tuple, Union

from ._impl import (_signature_channels,
                    _signature_forward,
                    _signature_backward)


# It would be lovely to do all of this at the C++ level. (In particular sigspec is really a struct that has no
# business being passed around at the Python level.) But unfortunately the documentation for how to create autograd
# Functions in C++ is nonexistent. Presumably that means it's all still subject to change, so we're just going to stick
# to the Python way of doings things for now.
class _SignatureFunction(autograd.Function):
    @staticmethod
    def forward(ctx: Any, path: torch.Tensor, depth: int, basepoint: bool = False, stream: bool = False,
                flatten: bool = True) -> Union[torch.Tensor, Tuple[torch.Tensor]]:
        result, result_as_vector, path_increments, sigspec = _signature_forward(path, depth, basepoint, stream, flatten)
        ctx._call_info = (result_as_vector, path_increments, sigspec, depth, basepoint, stream, flatten)
        if flatten:
            result = result[0]
        else:
            result = tuple(result)  # okay to return tuples, not okay to return lists. For some reason.
        return result

    @staticmethod
    def backward(ctx: Any, *grad_outputs: Tuple[torch.Tensor]) -> torch.Tensor:
        return _signature_backward(grad_outputs, *ctx._call_info), None, None, None, None


def signature(path: torch.Tensor, depth: int, basepoint: bool = False, stream: bool = False,
              flatten: bool = True) -> Union[torch.Tensor, Tuple[torch.Tensor]]:
    r"""Applies the signature transform to a stream of data.

    The input :attr:`path` is expected to be a three-dimensional tensor, with dimensions :math:`(N, C, L)`, where
    :math:`N` is the batch size, :math:`C` denotes the number of channels, and :math:`L` is the length of the input
    sequence. Thus each batch element is interpreted as a stream of data :math:`(x_1, \ldots, x_L)`, where each
    :math:`x_i \in \mathbb{R}^C`. (This is the same as ``torch.nn.Conv1D``, for example.)

    If :attr:`basepoint` is True then an additional point :math:`x_0 = 0 \in \mathbb{R}^C` is prepended to the path.

    Each path is then lifted to a piecewise constant path :math:`X \colon [0, 1] \to \mathbb{R}^C`, and the signature
    transform of this path is then computed. This (by the definition of the signature transform) gives a sequence of
    tensors of shape

    .. math::
        (C), (C, C), \ldots (C, \ldots, C),

    where the final tensor has :attr:`depth` many dimensions. If :attr:`flatten` is True then these are then flattened
    down and concatenated to give a single tensor of shape

    .. math::
        C + C^2 + \cdots + C^\text{depth}.

    (This value may be computed via the ``signatory.signature_channels`` function.)

    If :attr:`stream` is True then  the signatures of all intermediate paths :math:`(x_1, \ldots, x_j)`, for
    :math:`j=1, \ldots, L`, are also computed.

    Examples:
        If :attr:`stream` is False and :attr:`flatten` is True then the returned tensor will have shape

        .. math::
            (N, C + C^2 + \cdots + C^d).

        If :attr:`basepoint` is True, :attr:`stream` is True and :attr:`flatten` is True then the returned tensor will
        have shape

        .. math::
            (N, C + C^2 + \cdots + C^d, L),

        as the stream dimension is now preserved. See also the 'Returns' section below.

    Arguments:
        path (``torch.Tensor``): The input path to apply the signature transform to.

        depth (int): The depth to truncate the signature at.

        basepoint (bool, optional): Defaults to False. If True, then the path will have an additional point prepended to
            the start corresponding to the origin. If this is False then the signature transform is invariant to
            translations of the path. (Which may or may not be desirable.)

        stream (bool, optional): Defaults to False. If False then the signature transform of the whole path is computed.
            If True then the signature of all intermediate paths are also computed.

        flatten (bool, optional): Defaults to True. Whether to flatten the sequence of tensors resulting from the
            signature transform.

    Returns:
        A ``torch.Tensor`` or a tuple of ``torch.Tensor`` s.
        Given an input ``torch.Tensor`` of shape :math:`(N, C, L)`, and input arguments :attr:`depth`,
        :attr:`basepoint`, :attr:`stream`, :attr:`flatten`, then the return value is, in pseudocode:

        .. code-block:: python

            if flatten:
                if stream:
                    if basepoint:
                        return torch.Tensor of shape (N, C, L)
                    else:
                        return torch.Tensor of shape (N, C + C^2 + ... + C^(depth), L - 1)
                else:
                    return torch.Tensor of shape (N, C + C^2 + ... + C^(depth))
            else:
                if stream:
                    if basepoint:
                        out = []
                        for i in range(1, depth + 1):
                            out.append(torch.Tensor of shape (N, C^i, L))
                        return tuple(out)
                    else
                        out = []
                        for i in range(1, depth + 1):
                            out.append(torch.Tensor of shape (N, C^i, L - 1))
                        return tuple(out)
                else:
                    out = []
                    for i in range(1, depth + 1):
                        out.append(torch.Tensor of shape (N, C^i))
                    return tuple(out)

    """

    return _SignatureFunction.apply(path, depth, basepoint, stream, flatten)


# A wrapper for the sake of understandable documentation
def signature_channels(input_channels: int, depth: int):
    """Computes the number of output channels from a signature call.

    Arguments:
        input_channels (int): The number of channels in the input; that is, the dimension of the space that the input
            path resides in.

        depth (int): The depth of the signature that is being computed.

    Returns:
        An int specifying the number of channels in the signature of the path.
    """

    return _signature_channels(input_channels, depth)



import random
import signatory
import torch
import torch.autograd as autograd


def gradcheck(size=(1, 4, 3), depth=2, basepoint=False, stream=False, flatten=True):
    x = torch.rand(*size, requires_grad=True, dtype=torch.double)
    return autograd.gradcheck(signatory.signature, (x, depth, basepoint, stream, flatten))


# Fails, and it's not clear why. However...
def gradgradcheck(size=(1, 4, 3), depth=2, basepoint=False, stream=False, flatten=True):
    if flatten:
        test_fn = signatory.signature
    else:
        def test_fn(*args):
            result = signatory.signature(*args)
            return random.choice(result)
    x = torch.rand(*size, requires_grad=True, dtype=torch.double)
    return autograd.gradgradcheck(test_fn, (x, depth, basepoint, stream, flatten))


# ...this doesn't even manage to start at all, because of the fact that 'grad' and 'y' are both lists of torch.Tensors.
# Looking at the source code for autograd.gradcheck, it's clear that they don't expect any nested data structures for
# input like this; all their code assumems that the inputs are just torch.Tensors.
# Since we need to have lists of Tensors as input to signature_backward, it seems plausible that the fact that their
# torch.autograd.gradgradcheck even runs is a fluke, and the fact that it then fails is meaningless. Especially when
# you consider that we haven't even written a custom double-backward function - double-backward is done automatically by
# autograd in its usual 'tracing' manner, so it would be very surprising if it didn't give the correct results.
def gradgradcheck2(size=(1, 4, 3), depth=2, basepoint=False, stream=False, flatten=True):
    from signatory._impl import _signature_backward
    x = torch.rand(*size, dtype=torch.double)
    y = signatory.signature(x, depth, basepoint, stream, flatten)
    if flatten:
        # if not flatten then it's already a list
        y = [y]
    grad = [torch.rand_like(yi).requires_grad_() for yi in y]
    return autograd.gradcheck(_signature_backward, (grad, y, x, depth, basepoint, stream, flatten))



import io
import os
import setuptools
import torch.utils.cpp_extension as cpp

import metadata


ext_modules = [cpp.CppExtension(name='_impl', sources=['src/pytorchbind.cpp', 'src/signature.cpp'],
                                depends=['src/signature.hpp'])]

here = os.path.abspath(os.path.dirname(__file__))

with io.open(os.path.join(here, 'README.rst'), 'r', encoding='utf-8') as f:
    readme = f.read()

with io.open(os.path.join(here, 'docs', 'fragments', 'description.rst'), 'r', encoding='utf-8') as f:
    description = f.read()

setuptools.setup(name=metadata.project,
                 version=metadata.version,
                 author=metadata.author,
                 author_email=metadata.author_email,
                 maintainer=metadata.author,
                 maintainer_email=metadata.author_email,
                 description=description,
                 long_description=readme,
                 url=metadata.url,
                 license=license,
                 keywords=metadata.keywords,
                 classifiers=metadata.classifiers,
                 zip_safe=False,
                 python_requires=">=3.5",  # TODO: ">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, <4"
                 packages=[metadata.project],
                 ext_package=metadata.project,
                 package_dir={'': 'src'},
                 ext_modules=ext_modules,
                 cmdclass={'build_ext': cpp.BuildExtension})



import signatory
import iisignature
import timeit
import torch


def speed(batch_size=1000, number=10, depth=3, stream=True, backward=False):
    if stream and backward:
        raise ValueError("iisignature does not support backpropagation for a full stream of data, so we can't compare "
                         "against them.")
    signatory_x = torch.rand(batch_size, 10, 100, requires_grad=backward)
    iisignature_x = torch.rand(batch_size, 100, 10).numpy()

    def signatory_fn():
        y = signatory.signature(signatory_x, depth, stream=stream)
        if backward:
            y.backward(torch.rand_like(y))

    def iisignature_fn():
        for iisignature_ix in iisignature_x:
            y = iisignature.sig(iisignature_ix, depth, 2 if stream else 0)
            if backward:
                iisignature.sigbackprop(torch.rand_like(y), iisignature_ix, depth)

    signatory_time = timeit.timeit(signatory_fn, number=number)
    iisignature_time = timeit.timeit(iisignature_fn, number=number)
    print("signatory:   " + str(signatory_time))
    print("iisignature: " + str(iisignature_time))
    print("ratio:       " + str(signatory_time / iisignature_time))



import io
import os
import re


project = 'signatory'
author = "Patrick Kidger"
copyright = "2019, {}".format(author)
author_email = "contact@kidger.site"
url = "https://github.com/patrick-kidger/signatory"
license = "Apache-2.0"
keywords = "signature"
classifiers = ["Development Status :: 3 - Alpha",
               "Intended Audience :: Developers",
               "Intended Audience :: Financial and Insurance Industry",
               "Intended Audience :: Science/Research",
               "License :: OSI Approved :: Apache Software License",
               "Natural Language :: English",
               "Operating System :: Unix",               # TODO: test for:
                                                         #   Operating System :: Microsoft :: Windows
                                                         #   or
                                                         #   Operating System :: OS Independent
               "Programming Language :: Python :: 3",
               "Programming Language :: Python :: 3.5",
               "Programming Language :: Python :: 3.6",
               "Programming Language :: Python :: 3.7",  # TODO: test for:
                                                         #   Programming Language :: Python :: 2
               "Programming Language :: Python :: Implementation :: CPython",
               "Topic :: Scientific/Engineering :: Artificial Intelligence",
               "Topic :: Scientific/Engineering :: Information Analysis",
               "Topic :: Scientific/Engineering :: Mathematics"]

# for simplicity we actually store the version in the __version__ attribute in the source
with io.open(os.path.join(os.path.dirname(__file__), 'src', project, '__init__.py')) as f:
    meta_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", f.read(), re.M)
    if meta_match:
        version = meta_match.group(1)
    else:
        raise SystemExit("Unable to find __version__ string.")
